#!/usr/bin/perl -w
#
# gm.pl - maps overlay for Google Earth
#
# Copyright (C) 2006 Jiro Motonaga
# All rights reserved.
# This is free software with NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#
use strict;
use CGI;
#use CGI::Carp qw(fatalsToBrowser carpout);

my $pi  = 4 * atan2(1,1);
my $rd  = $pi / 180;

my $q = new CGI;
my $bbox = $q->param('BBOX');
my $v = $q->param('v');
my $r = $q->param('r') || 3;

#die 'Invalid arguments' if ($bbox =~ /[^0-9\,\.\-e]/);
#die 'Invalid range' unless ($r > 0);

my ($wlon, $slat, $elon, $nlat, $atlon, $atlat, $range, $tilt, $heading) = split /,/, $bbox;

my $zoom = int(log(80 * ($elon - $wlon)) / log(2) - $tilt / 90 + 0.5);
$zoom = 0 if ($zoom < 0);

$nlat = 85.05 if ($nlat > 85.05);
$slat = -85.05 if ($slat < -85.05);

my ($wx, $ny) = latlon2xy($nlat, $wlon, $zoom);
my ($ex, $sy) = latlon2xy($slat, $elon, $zoom);
my ($cx, $cy) = latlon2xy($atlat, $atlon, $zoom);

if ($ex - $wx > $r * 2) {
	$ex = $cx + $r;
	$wx = $cx - $r;
}
if ($sy - $ny > $r * 2) {
	$sy = $cy + $r;
	$ny = $cy - $r;
}

print <<KML;
Content-Type: application/vnd.google-earth.kml+xml

<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://earth.google.com/kml/2.0">
<Document>
<Folder>
<name>Google Maps</name>
<description>Google Maps (Zoom $zoom) <br />BBOX: $bbox</description>
KML

for (my $x = $wx; $x <= $ex; $x++) {
	for (my $y = $ny; $y <= $sy; $y++) {
		my ($north, $west) = xy2latlon($x, $y, $zoom);
		my ($south, $east) = xy2latlon($x+1, $y+1, $zoom);
		groundoverlay($v, $x, $y, $zoom, $north, $west, $south, $east);
	}
}

print <<KML;
</Folder>
</Document>
</kml>
KML

exit;

sub groundoverlay {
	my ($v, $x, $y, $zoom, $north, $west, $south, $east) = @_;
	if ($east > 180) {
		$west = -180 + ($west - 180);
		$east = -180 + ($east - 180);
		($x, $y) = latlon2xy($north, $west, $zoom);
	}
	my $url;
	if ($v eq 'kh') {
		my $quad = xy2quadtree($x, $y, $zoom);
		$url = "http://kh.google.com/kh?v=5&t=$quad";
	} else {
		$url = "http://mt.google.com/mt?v=$v&%78=$x&%79=$y&zoom=$zoom";
	}
	print <<KML;
	<GroundOverlay>
		<name>Google Map $x/$y/$zoom</name>
		<Icon>
			<href><![CDATA[$url]]></href>
		</Icon>
		<color>ffffffff</color>
		<LatLonBox>
			<north>$north</north>
			<south>$south</south>
			<east>$east</east>
			<west>$west</west>
			<rotation>0</rotation>
		</LatLonBox>
	</GroundOverlay>
KML

}

sub tan {
	sin($_[0]) / cos($_[0]);
}

sub latlon2xy {
	my ($lat, $lon, $zoom) = @_;
	my $x = 2**(17-$zoom) * (180 + $lon) / 360;
	my $y = 2**(17-$zoom) * ((log tan($pi/4 - $lat*$rd/2)) + $pi) / (2*$pi);
	return (int($x), int($y));
}

sub xy2latlon {
	my ($x, $y, $zoom) = @_;
	my $lon = ($x * 360) / (2**(17-$zoom)) - 180;
	my $tmp = (2*$pi*$y) / (2**(17-$zoom)) - $pi;
	my $lat = ($pi/4 - atan2(exp($tmp), 1)) * 2 / $rd;
	return ($lat, $lon);
}

sub xy2quadtree {
	my ($x, $y, $zoom) = @_;
	my @m = ('q', 't', 'r', 's');
	my @out;
	for (my $i = 0; $i < 17 - $zoom; $i++) {
		my $rx = $x % 2;
		my $ry = $y % 2;
		$x = int $x / 2;
		$y = int $y / 2;
		unshift @out, $m[oct("0b$rx$ry")];
	}
	return 't' . join '', @out;
}